@@ -1,4 +1,4 @@
-/* $Header: /p/tcsh/cvsroot/tcsh/sh.print.c,v 3.34 2011/01/24 18:10:26 christos Exp $ */
+/* $Header: /p/tcsh/cvsroot/tcsh/sh.print.c,v 3.35 2011/02/25 23:58:34 christos Exp $ */
 /*
  * sh.print.c: Primitive Output routines.
  */
@@ -32,7 +32,7 @@
  */
 #include "sh.h"
 
-RCSID("$tcsh: sh.print.c,v 3.34 2011/01/24 18:10:26 christos Exp $")
+RCSID("$tcsh: sh.print.c,v 3.35 2011/02/25 23:58:34 christos Exp $")
 
 #include "ed.h"
 
@@ -223,7 +223,6 @@ void
 flush(void)
 {
     int unit, oldexitset = exitset;
-    unsigned int errid = ERR_SILENT;
     static int interrupted = 0;
 
     /* int lmode; */
@@ -232,14 +231,16 @@ flush(void)
 	return;
     if (GettingInput && !Tty_raw_mode && linp < &linbuf[sizeof linbuf - 10])
 	return;
-    if (handle_intr) {
-	errid |= ERR_INTERRUPT;
+    if (handle_intr)
 	exitset = 1;
-    }
+
     if (interrupted) {
 	interrupted = 0;
 	linp = linbuf;		/* avoid recursion as stderror calls flush */
-	stderror(errid);
+	if (handle_intr)
+	    fixerror();
+	else
+	    stderror(ERR_SILENT);
     }
     interrupted = 1;
     if (haderr)
@@ -295,7 +296,10 @@ flush(void)
 		xexit(1);
 	    /*FALLTHROUGH*/
 	default:
-	    stderror(errid);
+	    if (handle_intr)
+		fixerror();
+	    else
+		stderror(ERR_SILENT);
 	    break;
 	}
 
