@@ -1,4 +1,4 @@
-/* $Header: /p/tcsh/cvsroot/tcsh/tc.prompt.c,v 3.68 2010/04/28 17:33:19 christos Exp $ */
+/* $Header: /p/tcsh/cvsroot/tcsh/tc.prompt.c,v 3.69 2010/05/17 19:26:42 christos Exp $ */
 /*
  * tc.prompt.c: Prompt printing stuff
  */
@@ -32,7 +32,7 @@
  */
 #include "sh.h"
 
-RCSID("$tcsh: tc.prompt.c,v 3.68 2010/04/28 17:33:19 christos Exp $")
+RCSID("$tcsh: tc.prompt.c,v 3.69 2010/05/17 19:26:42 christos Exp $")
 
 #include "ed.h"
 #include "tw.h"
@@ -213,8 +213,9 @@ tprintf(int what, const Char *fmt, const char *str, time_t tim, ptr_t info)
 		}
 		break;
 	    case '#':
-		Strbuf_append1(&buf,
-			       attributes | ((uid == 0 || euid == 0) ? PRCHROOT : PRCH));
+		Scp = (uid == 0 || euid == 0) ? PRCHROOT : PRCH;
+		if (Scp != '\0')
+		    Strbuf_append1(&buf, attributes | Scp);
 		break;
 	    case '!':
 	    case 'h':
