@@ -1,4 +1,4 @@
-/* $Header: /p/tcsh/cvsroot/tcsh/tc.str.c,v 3.38 2010/12/22 17:26:05 christos Exp $ */
+/* $Header: /p/tcsh/cvsroot/tcsh/tc.str.c,v 3.41 2011/02/27 00:15:17 christos Exp $ */
 /*
  * tc.str.c: Short string package
  * 	     This has been a lesson of how to write buggy code!
@@ -33,9 +33,10 @@
  */
 #include "sh.h"
 
+#include <assert.h>
 #include <limits.h>
 
-RCSID("$tcsh: tc.str.c,v 3.38 2010/12/22 17:26:05 christos Exp $")
+RCSID("$tcsh: tc.str.c,v 3.41 2011/02/27 00:15:17 christos Exp $")
 
 #define MALLOC_INCR	128
 #ifdef WIDE_STRINGS
@@ -392,17 +393,19 @@ s_strcasecmp(const Char *str1, const Char *str2)
 {
 #ifdef WIDE_STRINGS
     wint_t l1 = 0, l2 = 0;
-    for (; *str1 && ((*str1 == *str2 && (l1 = l2 = 0) == 0) || 
-	(l1 = towlower(*str1)) == (l2 = towlower(*str2))); str1++, str2++)
-	continue;
-    
+    for (; *str1; str1++, str2++)
+	if (*str1 == *str2)
+	    l1 = l2 = 0;
+	else if ((l1 = towlower(*str1)) != (l2 = towlower(*str2)))
+	    break;
 #else
     unsigned char c1, c2, l1 = 0, l2 = 0;
-    for (; *str1 && ((*str1 == *str2 && (l1 = l2 = 0) == 0) || 
-	((c1 = (unsigned char)*str1) == *str1 &&
-	 (c2 = (unsigned char)*str2) == *str2 &&
-	(l1 = tolower(c1)) == (l2 = tolower(c2)))); str1++, str2++)
-	continue;
+    for (; *str1; str1++, str2++)
+	if (*str1 == *str2)
+		l1 = l2 = 0;
+	else if ((l1 = tolower((unsigned char)*str1)) !=
+	    (l2 = tolower((unsigned char)*str2)))
+	    break;
 #endif
     /*
      * The following case analysis is necessary so that characters which look
@@ -625,6 +628,7 @@ STRBUF##_store1(struct STRBUF *buf, CHAR c)			\
 	    buf->size *= 2;					\
 	buf->s = xrealloc(buf->s, buf->size * sizeof(*buf->s));	\
     }								\
+    assert(buf->s);						\
     buf->s[buf->len] = c;					\
 }								\
 								\
